<?php

/**
 * This file contains PCSG\Api\Ebay
 */

namespace PCSG\Api;

use QUI;
use Stash\Driver\Apc;
use Stash\Driver\FileSystem;
use Stash\Pool;

/**
 * PCSG eBay API
 * Mini API access
 *
 * @author www.pcsg.de (Henning Leutz)
 * @author www.pcsg.de (Patrick Müller)
 *
 * API Status:
 * https://go.developer.ebay.com/api-status
 *
 * Release Notes:
 * http://developer.ebay.com/devzone/xml/docs/releasenotes.html#957
 */
class Ebay extends QUI\QDOM
{
    /**
     * Internal Session-ID
     *
     * @var String
     */
    protected $Session;

    /** @var string URL to the eBay FindingService API endpoint */
    const URL_FINDING_SERVICE = "https://svcs.ebay.com/services/search/FindingService/v1";

    /**
     * Constructor
     * @param array $attributes - optional
     * @throws \Exception
     */
    public function __construct($attributes)
    {
        // defaults
        $this->setAttributes([
            'SECURITY-APPNAME'     => '',
            'GLOBAL-ID'            => 'EBAY-DE',
            'OPERATION-NAME'       => 'findItemsAdvanced',
            'SERVICE-VERSION'      => '1.11.0',
            'REST-PAYLOAD'         => '',
            'affiliate.networkId'  => 9,
            'affiliate.trackingId' => '5336210971',
            'affiliate.customId'   => '5336210971',
            'sandbox'              => false,
            'debug'                => false,
            // for authentication
            'DEV-NAME'             => '',
            'CERT-NAME'            => '',
            'COMPATIBILITY-LEVEL'  => 681,
            'CALL-NAME'            => 'FetchToken',
            'RuName'               => ''
        ]);

        // set options
        $this->setAttributes($attributes);

        // check app id
        if (!$this->getAttribute('SECURITY-APPNAME')) {
            throw new \Exception(
                'No Access. No eBay SECURITY-APPNAME given',
                401
            );
        }
    }

    /**
     * Make a normal eBay request with the set attributes
     *
     * @return array
     */
    public function search()
    {
        $attr = $this->getFindingServiceAttributes();

        $url = self::URL_FINDING_SERVICE . '?' . http_build_query($attr);

        $data = $this->executeGETRequest($url);
        $data = json_decode($data, true);

        return $this->parseItems($data);
    }

    /**
     * Return the Articles from the seller
     *
     * @param string|boolean $sellerid - seller id, seller name
     * @return array
     */
    public function getItemsBySeller($sellerid = false, $noCache = false)
    {
        $attr = $this->getFindingServiceAttributes();

        $attr['itemFilter(0).name']     = 'Seller';
        $attr['itemFilter(0).value(0)'] = $sellerid;
        $attr['itemFilter(1).name']     = 'HideDuplicateItems';
        $attr['itemFilter(1).value(1)'] = '1';
        $attr['outputSelector']         = 'UnitPriceInfo';  // get amount of and unit article e.g: "500g"

        $parts = [];

        foreach ($attr as $key => $value) {
            $parts[] = sprintf('%s=%s', $key, rawurlencode($value));
        }

        $pool = $this->getStash();
        $Item = $pool->getItem('itemsBySeller/' . $sellerid . '/' . md5(serialize($parts)));
        $data = $Item->get();

        if ($noCache || $Item->isMiss()) {
            $url = self::URL_FINDING_SERVICE . '?' . implode('&', $parts);

            $data = $this->executeGETRequest($url);
            $data = json_decode($data, true);
            $data = $this->parseItems($data);

            $Item->set($data, 1800);
        }

        return $data;
    }


    /**
     * Returns the number of Articles from a seller
     *
     * @param string|boolean $sellerID - seller id, seller name
     * @return integer
     */
    public function getNumberItemsBySeller($sellerID = false)
    {
        $pool = $this->getStash();
        $item = $pool->getItem('numberItemsBySellerId/' . $sellerID);
        $data = $item->get();

        if ($item->isMiss()) {
            $attr = $this->getFindingServiceAttributes();

            $attr['itemFilter(0).name']     = 'Seller';
            $attr['itemFilter(0).value(0)'] = $sellerID;
            $attr['itemFilter(1).name']     = 'HideDuplicateItems';
            $attr['itemFilter(1).value(1)'] = '1';

            unset($attr['paginationInput.entriesPerPage']);

            $parts = [];

            foreach ($attr as $key => $value) {
                $parts[] = sprintf('%s=%s', $key, rawurlencode($value));
            }

            $get = self::executeGETRequest(self::URL_FINDING_SERVICE . '?' . implode('&', $parts));

            $data = json_decode($get, true);

            if (isset($data['findItemsAdvancedResponse'][0]['errorMessage']) || isset($data['errorMessage'][0]['error'])) {
                $data = 0;
            } else {
                $data = $data['findItemsAdvancedResponse'][0]['paginationOutput'][0]['totalEntries'][0];
                $data = intval($data);
            }

            $item->set($data, 1800);
        }

        return $data;
    }


    /**
     * Returns information about an article
     *
     * @param $articleId
     * @return array
     */
    public function getItemById($articleId)
    {
        $sideId = $this->globalIDToSiteID($this->getAttribute('GLOBAL-ID'));

        $pool = $this->getStash();
        $item = $pool->getItem('itemsById/' . $articleId . '/' . $sideId);
        $data = $item->get();

        if (true || $item->isMiss()) {
            $item->lock();

            $attr = $this->getFindingServiceAttributes();

            $url = 'http://open.api.ebay.com/shopping?';
            $url .= '&callname=GetSingleItem';
            $url .= '&responseencoding=JSON';
            $url .= '&appid=' . $attr['SECURITY-APPNAME'];
//            $url .= '&siteid=0';
            $url .= '&version=967';
            $url .= '&IncludeSelector=ItemSpecifics,Details';
            $url .= '&trackingpartnercode=' . $attr['affiliate.networkId'];
            $url .= '&trackingid=' . $attr['affiliate.trackingId'];
            $url .= '&affiliateuserid=' . $attr['affiliate.customId'];
            $url .= '&ItemID=' . $articleId;
            $url .= '&siteid=' . $sideId;

            $get  = self::executeGETRequest($url);
            $data = json_decode($get, true);

            // if article not found, cache error forever
            if (isset($data['Errors'])) {
                $item->set($data);
            } else {
                $item->set($data, 600);
            }
        }

        return $data;
    }

    /**
     * Get number of results for a specific search term
     *
     * @param string $searchStr - Search string
     * @return int
     */
    public function getItemSearchResultCount($searchStr)
    {
        $params = [
            'OPERATION-NAME'                 => 'findItemsByKeywords',
            'SERVICE-VERSION'                => $this->getAttribute('SERVICE-VERSION'),
            'SECURITY-APPNAME'               => $this->getAttribute('SECURITY-APPNAME'),
            'GLOBAL-ID'                      => $this->getAttribute('GLOBAL-ID'),
            'keywords'                       => $searchStr,
            'paginationInput.entriesPerPage' => 1,
            'RESPONSE-DATA-FORMAT'           => 'JSON',
            'affiliate.networkId'            => $this->getAttribute('affiliate.networkId'),
            'affiliate.trackingId'           => $this->getAttribute('affiliate.trackingId')
        ];

        $url = self::URL_FINDING_SERVICE . '?' . http_build_query($params);

        $data = $this->executeGETRequest($url);
        $data = json_decode($data, true);

        if (empty($data['findItemsByKeywordsResponse'])) {
            return false;
        }

        $data = $data['findItemsByKeywordsResponse'][0];

        if (!isset($data['ack']) || $data['ack'][0] !== 'Success') {
            return false;
        }

        $this->setAttribute('itemSearchURL', $data['itemSearchURL'][0]);
        $count = $data['paginationOutput'][0]['totalEntries'][0];

        return $count;
    }

    /**
     * Check if an eBay username is taken
     *
     * @param string $username
     * @return bool
     */
    public function existsUsername($username)
    {
        $globalId = $this->getAttribute('GLOBAL-ID');

        $data = $this->executeGETRequest(
            'http://open.api.ebay.com/shopping?' . http_build_query([
                'callname'         => 'FindItemsAdvanced',
                'version'          => '601',
                'appid'            => $this->getAttribute('SECURITY-APPNAME'),
                'siteid'           => $this->globalIDToSiteID($globalId),
                'SellerID'         => $username,
                'responseencoding' => 'JSON'
            ])
        );

        $data = json_decode($data, true);

        if (isset($data['Ack']) && $data['Ack'] == 'Success') {
            return true;
        }

        return false;
    }

    /**
     * Return the
     *
     * @return array
     */
    protected function getFindingServiceAttributes()
    {
        $attr = $this->getAttributes();

        // old api compatiblity
        $attr['RESPONSE-DATA-FORMAT'] = 'JSON';

        if (!isset($attr['paginationInput.entriesPerPage'])) {
            $attr['paginationInput.entriesPerPage'] = 10;
        }

        if (!isset($attr['outputSelector'])) {
            $attr['outputSelector'] = 'PictureURLLarge';
        }

        if (!isset($attr['keywords'])) {
            $attr['keywords'] = '';
        }

        return $attr;
    }

    /**
     * Parse the return of a FindingService request
     *
     * @param array $data
     * @return array
     * @throws \Exception
     */
    protected function parseItems($data)
    {
        if (isset($data['findItemsAdvancedResponse'])) {
            $_data = $data['findItemsAdvancedResponse'];
        } else {
            if (isset($data['findItemsByKeywordsResponse'])) {
                $_data = $data['findItemsByKeywordsResponse'];
            }
        }


        if (!isset($_data[0])) {
            return [];
        }

        if (isset($_data[0]['ack']) &&
            $_data[0]['ack'][0] == 'Failure'
        ) {
            throw new \Exception(
                $_data[0]['errorMessage'][0]['error'][0]['message'][0]
            );
        }

        if (!isset($_data[0]['searchResult']) ||
            !isset($_data[0]['searchResult'][0])
        ) {
            return [];
        }

        if ($_data[0]['searchResult'][0]['@count'] == 0) {
            return [];
        }

        $result = [];
        $list   = $_data[0]['searchResult'][0]['item'];

        foreach ($list as $item) {
            if (!isset($item['galleryURL'])) {
                $item['galleryURL'][0] = '';
            }

            if (!isset($item['pictureURLLarge'])) {
                $item['pictureURLLarge'][0] = $item['galleryURL'][0];
            }

            if (!isset($item['shippingInfo']) ||
                !isset($item['shippingInfo'][0]['shippingServiceCost']) ||
                !isset($item['shippingInfo'][0]['shippingServiceCost'][0]['__value__'])
            ) {
                $item['shippingInfo'][0]['shippingServiceCost'][0]['__value__'] = '';
            }

            if (!isset($item['shippingInfo']) ||
                !isset($item['shippingInfo'][0]['shippingServiceCost']) ||
                !isset($item['shippingInfo'][0]['shippingServiceCost'][0]['@currencyId'])
            ) {
                $item['shippingInfo'][0]['shippingServiceCost'][0]['@currencyId'] = '';
            }

            if (!isset($item['sellingStatus']) ||
                !isset($item['sellingStatus'][0]['convertedCurrentPrice']) ||
                !isset($item['sellingStatus'][0]['convertedCurrentPrice'][0]['__value__'])
            ) {
                $item['sellingStatus'][0]['convertedCurrentPrice'][0]['__value__'] = '';
            }

            if (!isset($item['sellingStatus']) ||
                !isset($item['sellingStatus'][0]['convertedCurrentPrice']) ||
                !isset($item['sellingStatus'][0]['convertedCurrentPrice'][0]['@currencyId'])
            ) {
                $item['sellingStatus'][0]['convertedCurrentPrice'][0]['@currencyId'] = '';
            }

            $title = $item['title'][0];

            if (isset($item['eekStatus'])) {
                $title = $title . ' | EEK ' . $item['eekStatus'][0];
            }


            $price = [
                'value'    => QUI\Utils\Convert::formPrice(
                    $item['sellingStatus'][0]['convertedCurrentPrice'][0]['__value__'],
                    3
                ),
                'currency' => $item['sellingStatus'][0]['convertedCurrentPrice'][0]['@currencyId']
            ];

            $unitPrice = false;

            // Calculate price per 100g/1kg/1 piece etc.
            if (isset($item['unitPrice'])) {
                $amount = $item['unitPrice'][0]['quantity'][0];

                $unitValue = QUI\Utils\Convert::formPrice(($price['value'] / $amount), 3);

                $unitPrice['value'] = $unitValue;
                $unitPrice['unit']  = $item['unitPrice'][0]['type'][0];
            }

            $result[] = [
                'itemId'              => $item['itemId'][0],
                'title'               => $title,
                'endTime'             => $item['listingInfo'][0]['endTime'][0],
                'url'                 => $item['viewItemURL'][0],
                'listingType'         => $item['listingInfo'][0]['listingType'],
                'imageUrl'            => str_replace('http://', 'https://', $item['pictureURLLarge'][0]),
                'galleryURL'          => str_replace('http://', 'https://', $item['galleryURL'][0]),
                'primaryCategoryId'   => $item['primaryCategory'][0]['categoryId'][0],
                'primaryCategoryName' => $item['primaryCategory'][0]['categoryName'][0],
                'price'               => $price,
                'unitPrice'           => $unitPrice,
                'timeLeft'            => $this->getPrettyTimeFromEbayTime(
                    $item['sellingStatus'][0]['timeLeft'][0]
                ),
                'shippingCost'        => [
                    'value'    => QUI\Utils\Convert::formPrice(
                        $item['shippingInfo'][0]['shippingServiceCost'][0]['__value__'],
                        3
                    ),
                    'currency' => $item['shippingInfo'][0]['shippingServiceCost'][0]['@currencyId']
                ]
            ];
        }

        return $result;
    }

    /**
     * Convert the eBay timestamp to a pretty format
     *
     * @param String $eBayTimeString Input is of form 'PT12M25S'
     * @return string
     * @todo time string muss nicht unbedingt in cache
     */
    public static function getPrettyTimeFromEbayTime($eBayTimeString)
    {
        $pool = self::getStash();
        $item = $pool->getItem('prettyTimeFromEbayTime/' . $eBayTimeString);
        $data = $item->get();

        if ($item->isMiss()) {
            if ($eBayTimeString == 'PT0S') {
                $data = "Auktion beendet.";
                $item->set($data);

                return $data;
            }

            $matchAry = [];
            $pattern  = "#P([0-9]{0,3}D)?T([0-9]?[0-9]H)?([0-9]?[0-9]M)?([0-9]?[0-9]S)#msiU";
            preg_match($pattern, $eBayTimeString, $matchAry);

            $days  = false;
            $hours = false;
            $min   = false;
            $sec   = false;

            if (isset($matchAry[1])) {
                $days = (int)$matchAry[1];
            }

            if (isset($matchAry[2])) {
                $hours = (int)$matchAry[2];
            }

            if (isset($matchAry[3])) {
                $min = (int)$matchAry[3];
            }

            if (isset($matchAry[4])) {
                $sec = (int)$matchAry[4];
            }


            $retnStr = '';

            if ($days) {
                $retnStr .= "$days Tag(e),";
            }

            if ($hours) {
                $retnStr .= " $hours Stunde(n),";
            }

            if ($min) {
                $retnStr .= " $min Minute(n),";
            }

            if ($sec) {
                $retnStr .= " $sec Sekunde(n)";
            }

            $data = $retnStr;
            $item->set($data, 2419200); // ablauf datum auf 4 wochen
        }

        return $data;

        /* ISO 8601 duration specification and can be use with DateTime() and DateInterval()
            Braucht allerdings länger (0.00011515617370605s) als Methode oben (3.9100646972656E-5s)
        $date = new \DateTime();
        $date->add(new \DateInterval($eBayTimeString));
        $now = new \DateTime();

        $interval = $date->diff($now);

        $format = "";
        if($interval->d) $format .= "%d Tag(e), ";
        if($interval->h) $format .= "%h Stunde(n), ";
        if($interval->i) $format .= "%i Minute(n), ";
        if($interval->s) $format .= "%s Sekunde(n), ";
        return $interval->format($format);
        */
    }

    /**
     * If a session already exist you can set the internal Session-ID
     *
     * @param String $SessID
     */
    public function setSession($SessID)
    {
        $this->Session = $SessID;
    }

    /**
     * Create a ebay session, if a session is set, this session would be return
     *
     * @return String
     */
    public function getSession()
    {
        if (!empty($this->Session)) {
            return $this->Session;
        }

        // get session id
        $this->setAttribute('CALL-NAME', 'GetSessionID');

        $Response = $this->sendHttpRequest(
            '<?xml version="1.0" encoding="utf-8" ?>' .
            '<GetSessionIDRequest xmlns="urn:ebay:apis:eBLBaseComponents">' .
            '<Version>' . $this->getAttribute('COMPATIBILITY-LEVEL') . '</Version>' .
            '<RuName>' . $this->getAttribute('RuName') . '</RuName>' .
            '</GetSessionIDRequest>'
        );

        $this->debug($Response);


        $ResponseXML = simplexml_load_string($Response);

        $this->setSession((string)$ResponseXML->SessionID);

        return $this->Session;
    }

    /**
     * Ask for a token
     */
    public function getToken()
    {
        $this->setAttribute('CALL-NAME', 'FetchToken');

        $response = $this->sendHttpRequest(
            '<?xml version="1.0" encoding="utf-8" ?>' .
            '<FetchTokenRequest xmlns="urn:ebay:apis:eBLBaseComponents">' .
            '<SessionID>' . $this->getSession() . '</SessionID>' .
            '</FetchTokenRequest>'
        );

        $this->debug($response);


        $ResponseXML = simplexml_load_string($response);

        // if login not allowed
        // then ask for it
        if ($ResponseXML->Ack == 'Failure') {
            throw new \Exception('Get Token not allowed');
        }

        $Token      = (string)$ResponseXML->eBayAuthToken;
        $expiration = $ResponseXML->HardExpirationTime;

        return $Token;
    }

    /**
     * Fetch the signin form for the session
     *
     * @return String
     */
    public function signInForm()
    {
        return file_get_contents($this->signInUrl());
    }

    /**
     * Return the signin Url
     *
     * @return string
     */
    public function signInUrl()
    {
        $url = 'https://signin.ebay.com/ws/eBayISAPI.dll?';

        if ($this->getAttribute('sandbox')) {
            $url = 'https://signin.sandbox.ebay.com/ws/eBayISAPI.dll?';
        }

        $attr = [
            'RuName' => $this->getAttribute('RuName'),
            'SessID' => $this->getSession()
        ];

        return $url . http_build_query($attr);
    }

    /**
     * Return the difference between two dates
     *
     * @param string $start - start date, ebay date
     * @param string|boolean $end - end date, ebay date
     * @return String
     */
    public function dateDiff($start, $end = false)
    {
        try {
            $start = new \DateTime($start);
            $end   = new \DateTime($end);

            $form = $start->diff($end);
        } catch (\Exception $Exception) {
            return $Exception->getMessage();
        }

        $hour = sprintf("%02d", $form->h);
        $min  = sprintf("%02d", $form->i);
        $sec  = sprintf("%02d", $form->s);

        $year  = sprintf("%02d", $form->y);
        $month = sprintf("%02d", $form->m);
        $days  = sprintf("%02d", $form->d);

        $result = '';

        if ($form->y || $form->m || $form->d) {
            if (!$form->y || !$form->m || $form->d) {
                $result = "{$days} Tage, ";
            } else {
                $result = "{$days} Tage, {$month} Monate, {$year} Jahre ";
            }
        }

        $result .= "{$hour}:{$min}:{$sec}";

        return $result;
    }

    /**
     * Build the ebay request header for authentication
     *
     * @return array
     */
    protected function buildHeaders()
    {
        return [
            // Regulates versioning of the XML interface for the API
            'X-EBAY-API-COMPATIBILITY-LEVEL: ' . $this->getAttribute('COMPATIBILITY-LEVEL'),
            //set the keys
            'X-EBAY-API-DEV-NAME: ' . $this->getAttribute('DEV-NAME'),
            'X-EBAY-API-APP-NAME: ' . $this->getAttribute('SECURITY-APPNAME'),
            'X-EBAY-API-CERT-NAME: ' . $this->getAttribute('CERT-NAME'),
            'X-EBAY-API-CALL-NAME: ' . $this->getAttribute('CALL-NAME'),
            'X-EBAY-API-SITEID: ' . $this->globalIDToSiteID(
                $this->getAttribute('GLOBAL-ID')
            )
        ];
    }

    /**
     * Return the eBay Site-ID of the eBay Global-ID
     *
     * @param string $globalID
     * @return integer
     */
    protected function globalIDToSiteID($globalID)
    {
        $ids = [
            'EBAY-AT'    => 16,
            'EBAY-AU'    => 15,
            'EBAY-CH'    => 193,
            'EBAY-DE'    => 77,
            'EBAY-ENCA'  => 2,
            'EBAY-ES'    => 186,
            'EBAY-FR'    => 71,
            'EBAY-FRBE'  => 23,
            'EBAY-FRCA'  => 210,
            'EBAY-GB'    => 3,
            'EBAY-HK'    => 201,
            'EBAY-IE'    => 205,
            'EBAY-IN'    => 203,
            'EBAY-IT'    => 101,
            'EBAY-MOTOR' => 100,
            'EBAY-MY'    => 207,
            'EBAY-NL'    => 146,
            'EBAY-NLBE'  => 123,
            'EBAY-PH'    => 211,
            'EBAY-PL'    => 212,
            'EBAY-SG'    => 216,
            'EBAY-US'    => 0
        ];

        return isset($ids[$globalID]) ? $ids[$globalID] : 77;
    }

    /**
     * Send an http request to eBay
     *
     * @param String $requestBody - XML request
     * @return mixed
     */
    protected function sendHttpRequest($requestBody)
    {
        $serverUrl = 'https://api.ebay.com/ws/api.dll';
//        $shoppingURL = 'http://open.api.ebay.com/shopping';
//        $findingURL  = 'http://svcs.ebay.com/services/search/FindingService/v1';

        if ($this->getAttribute('sandbox')) {
            $serverUrl = 'https://api.sandbox.ebay.com/ws/api.dll';
//            $shoppingURL = 'http://open.api.sandbox.ebay.com/shopping';
//            $findingURL  = 'http://svcs.sandbox.ebay.com/services/search/FindingService/v1';
        }

        $Curl = curl_init();

        curl_setopt($Curl, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($Curl, CURLOPT_SSL_VERIFYHOST, 0);

        curl_setopt($Curl, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($Curl, CURLOPT_POST, 1);

        curl_setopt($Curl, CURLOPT_URL, $serverUrl);
        curl_setopt($Curl, CURLOPT_HTTPHEADER, $this->buildHeaders());
        curl_setopt($Curl, CURLOPT_POSTFIELDS, $requestBody);

        $Response = curl_exec($Curl);

        curl_close($Curl);

        return $Response;
    }

    /**
     * Execute a GET request via cURL
     *
     * @param string $url
     * @return string|false
     */
    protected function executeGETRequest($url)
    {
        $Curl = curl_init($url);

        curl_setopt($Curl, CURLOPT_RETURNTRANSFER, true);

        curl_setopt($Curl, CURLOPT_SSL_VERIFYPEER, 0);
        curl_setopt($Curl, CURLOPT_SSL_VERIFYHOST, 0);

        curl_setopt($Curl, CURLOPT_CONNECTTIMEOUT, 10);
        curl_setopt($Curl, CURLOPT_TIMEOUT, 10);

        $response = curl_exec($Curl);

        curl_close($Curl);

        return $response;
    }

    /**
     * Log a message via error_log
     * only if debug is true
     *
     * @param String $message - the messsage that would be log
     */
    public function debug($message)
    {
        if ($this->getAttribute('debug')) {
            error_log("\n\n" . $message);
        }
    }

    /**
     * @return Pool
     */
    public static function getStash()
    {
        if (class_exists('\QUI\Cache\Manager')) {
            QUI\Cache\Manager::getStash();

            return QUI\Cache\Manager::$Stash;
        }

        if (Apc::isAvailable()) {
            $driver = new Apc();

            return new Pool($driver);
        }

        $driver = new FileSystem();

        return new Pool($driver);
    }
}
